

#include <string.h>

#ifdef WIN32
#include <windows.h>
#include <GL/gl.h>
#endif

#include "timer.h"
#include "texture.h"
#include "gfx.h"

#include "vgev_tile.h"


/* ************************************************************
************************************************************ */
class point_node : public dbl_llist {

   public:
      vector2i pt;
      virtual ~point_node() {}
};


/* ************************************************************
************************************************************ */
class tile_node : public dbl_llist {

   public:
      vector2i ll, ur;
      texbase *tob;
      virtual ~tile_node() {}
};


/* ************************************************************
************************************************************ */
int image_block_type::parse(FILE *infile, char *token) {

   get_token(infile, token);
   basename.stringcpy(token);
   
   get_token(infile, token);
   subcount = atoi(token);

   return 1;
}


/* ************************************************************
************************************************************ */
int image_block_type::preprocess() {

   FILE *infile;
   char buffer[MAXSTRLEN], buffer2[MAXSTRLEN];
   int i, j, m;
   sub_image_block_type *sbtr;
   int row, col;
   double neo, nea, swo, swa, dlat, dlon;

   for (i=0; i<subcount; i++) {

      sprintf(buffer, "%s%d", basename.string, i);
      if (!find_file(buffer, "r", "", PLATFORM_SLASH, NULL, &infile))
         return 0;

      while (1) {
         if (fscanf(infile, "%s %d %d %lf %lf %lf %lf %lf %lf", buffer, &row, &col, &swa, &swo, &nea, &neo, &dlat, &dlon) != 9)
            break;

         for (sbtr = (sub_image_block_type *)subblockman.head; sbtr && sbtr->basename.stringcmp(buffer); sbtr=(sub_image_block_type *)sbtr->next);

         if (!sbtr) {
            if (i) {
               fclose(infile);
               return 0;
            }

            subblockman.insert(sbtr = new sub_image_block_type, NULL);

            sbtr->basename.stringcpy(buffer);
            sbtr->swa = swa;
            sbtr->swo = swo;
            sbtr->nea = nea;
            sbtr->neo = neo;
            sbtr->count = subcount;
            sbtr->current_level = sbtr->count-1;
            sbtr->tile_level = new image_tile_level_type[sbtr->count];
         }

         sbtr->tile_level[i].dlat = dlat;
         sbtr->tile_level[i].dlon = dlon;
         sbtr->tile_level[i].count = row*col;
         sbtr->tile_level[i].tile = new image_tile_type[sbtr->tile_level[i].count];

         strcpy(buffer2, basename.string);
         for (m=strlen(buffer2)-1; m>-1; m--)
            if (buffer2[m] == '/' || buffer2[m] == '\\' || buffer2[m] == ':') {
               buffer2[m+1] = 0;
               break;
            }

         m = strlen(buffer2);
         if (buffer2[m-1] != '/' && buffer2[m-1] != '\\'  && buffer2[m-1] != ':') {
            buffer2[m] = PLATFORM_SLASH;
            buffer2[m+1] = 0;
         }

         for (j=0; j<sbtr->tile_level[i].count; j++) {
            fscanf(infile, "%s %lf %lf %lf %lf", buffer, &sbtr->tile_level[i].tile[j].swlat, &sbtr->tile_level[i].tile[j].swlon, &sbtr->tile_level[i].tile[j].nelat, &sbtr->tile_level[i].tile[j].nelon);
            sbtr->tile_level[i].tile[j].filename.stringcpy(buffer2);
            sbtr->tile_level[i].tile[j].filename.stringcat(buffer);
         }

      }

      fclose(infile);
   }

   return 1;
}


/* ************************************************************
************************************************************ */
int lod_type::parse(FILE *infile, char *token) {

   int j;
   
   if (blocks)
      delete [] blocks;
      
   get_token(infile, token);
   swap_up = atof(token);

   get_token(infile, token);
   swap_down = atof(token);

   get_token(infile, token);
   count = atoi(token);

   blocks = new image_block_type[count];

   for (j=0; j<count; j++)
      blocks[j].parse(infile, token);

   return 1;
}


/* ************************************************************
************************************************************ */
int lod_type::preprocess() {

   int i;
   
   for (i=0; i<count; i++)
      if (!blocks[i].preprocess())
         return 0;

   return 1;
}


/* ************************************************************
************************************************************ */
void tile_manager_type::dest() {

   lod_count = 0;

   if (lods) {
      delete [] lods;
      lods = NULL;
   }

}


/* ************************************************************
************************************************************ */
int tile_manager_type::read_tiles(char *filename) {

   FILE *infile;
   char buffer[MAXSTRLEN];
   int i;

   dest();
   
   if (!find_file(filename, "r", "", PLATFORM_SLASH, NULL, &infile))
      return 0;

   get_token(infile, buffer);
   lod_count = atoi(buffer);

   if (!lod_count) {
      fclose(infile);
      return 0;
   }
   
   lods = new lod_type[lod_count];
   lod_current_level = lod_count-1;

   for (i=0; i<lod_count; i++)
      lods[i].parse(infile, buffer);

   fclose(infile);

   for (i=0; i<lod_count; i++)
      if (!lods[i].preprocess()) {
         dest();
         return 0;
      }

   return 1;
}


/* ************************************************************
************************************************************ */
void tile_manager_type::preprocess_map(canvas *tgev) {

   double factor, current;
   double ix = 1.0/WINX;
   int i, j, k, l;
   int llx, lly, urx, ury;
   sub_image_block_type *sbtr;
   image_tile_type *ptr;

   tile_node *ttr;
   point_node *pointtr;

   pointman.dest();
   visman.dest();

   if (!lods || !tgev)
      return;

   factor = 1.0/tgev->get_scale();

   // calculate current lod
   current = lods[lod_current_level].swap_up*ix;

   if (factor < current) {
      while (lod_current_level && factor < current) {
         lod_current_level--;
         current = lods[lod_current_level].swap_up*ix;
      }

      for (i=0; i<lods[lod_current_level].count; i++)
         for (sbtr = (sub_image_block_type *)lods[lod_current_level].blocks[i].subblockman.head; sbtr; sbtr = (sub_image_block_type *)sbtr->next)
            sbtr->current_level = sbtr->count-1;
   }

   else {
      current = lods[lod_current_level].swap_down*ix;

      if (factor > current) {
         while (lod_current_level < lod_count-1 && factor > current) {
            lod_current_level++;
            current = lods[lod_current_level].swap_down*ix;
         }

         for (i=0; i<lods[lod_current_level].count; i++)
            for (sbtr = (sub_image_block_type *)lods[lod_current_level].blocks[i].subblockman.head; sbtr; sbtr = (sub_image_block_type *)sbtr->next)
               sbtr->current_level = sbtr->count-1;
      }

   }

   for (i=0; i<lods[lod_current_level].count; i++)
      for (sbtr = (sub_image_block_type *)lods[lod_current_level].blocks[i].subblockman.head; sbtr; sbtr = (sub_image_block_type *)sbtr->next) {
         current = sbtr->tile_level[sbtr->current_level].dlat*0.75*0.95;

         if (factor < current) {
            while (sbtr->current_level && factor < current) {
               sbtr->current_level--;
               current = sbtr->tile_level[sbtr->current_level].dlat*0.75*0.95;
            }

         }

         else {
            current = sbtr->tile_level[sbtr->current_level].dlat*1.5*1.05;

            while (sbtr->current_level < sbtr->count-1 && factor > current) {
               sbtr->current_level++;
               current = sbtr->tile_level[sbtr->current_level].dlat*1.5*1.05;
            }

         } 

         for (j=0; j<sbtr->tile_level[sbtr->current_level].count; j++) {
            ptr = &sbtr->tile_level[sbtr->current_level].tile[j];
            k = tgev->clip(ptr->swlon, ptr->swlat, &llx, &lly);
            l = tgev->clip(ptr->nelon, ptr->nelat, &urx, &ury);

            if ((k & l) || llx == urx || lly == ury) {
               pointman.append(pointtr = new point_node, NULL);
               pointtr->pt[0] = llx;
               pointtr->pt[1] = lly;
               continue;
            }

            if (!ptr->tob) {
               if (ptr->initflag)
                  continue;

               ptr->tob = (texbase *)((frame_manager *)global_resource_manager)->read_tex(ptr->filename.string);
               if (!ptr->tob)
                  continue;
            }

            visman.append(ttr = new tile_node, NULL);
            ttr->tob = ptr->tob;
            ttr->ll[0] = llx;
            ttr->ll[1] = lly;
            ttr->ur[0] = urx;
            ttr->ur[1] = ury;
          }

      }

}


/* ************************************************************
************************************************************ */
void tile_manager_type::render(canvas *tgev) {

   tile_node *ttr;
   point_node *pointtr;
   texture_block *tblock;
   TIME_STRUCT t;

   tgev->tpColor4ubv(tile_color, 0);

   glBegin(GL_POINTS);
      for (pointtr = (point_node *)pointman.head; pointtr; pointtr = (point_node *)pointtr->next)
         glVertex2iv(pointtr->pt);
   glEnd();

   glEnable(GL_TEXTURE_2D);
   gfxEnable(GFX_INVISO);

   for (ttr = (tile_node *)visman.head; ttr; ttr=(tile_node *)ttr->next) {
      tblock = (texture_block *)ttr->tob->query_data();
      glBindTexture(GL_TEXTURE_2D, tblock->user_id);

      glBegin(GL_POLYGON);
         glTexCoord2f(0,0);
         glVertex2iv(ttr->ll);
         glTexCoord2f(0,1.0f);
         glVertex2i(ttr->ll[0], ttr->ur[1]);
         glTexCoord2f(1.0f,1.0f);
         glVertex2iv(ttr->ur);
         glTexCoord2f(1.0f,0);
         glVertex2i(ttr->ur[0], ttr->ll[1]);
      glEnd();

   }

   gfxDisable(GFX_INVISO);
   glDisable(GL_TEXTURE_2D);

   GETTIME(&t);
   global_resource_manager->update((float)TIME2FLOAT(t));
}


